

#ifndef __WANPIPE_KERNEL_H
# define __WANPIPE_KERNEL_H


#if defined(WAN_KERNEL)

#if defined(__FreeBSD__)
/************************* F R E E B S D *************************************/
# if  (__FreeBSD_version < 600000)
#  define IFF_DRV_OACTIVE		IFF_OACTIVE
#  define IFF_DRV_RUNNING		IFF_RUNNING
# endif
# if defined(IFP2AC)
#  define WAN_IFP2AC(ifp)		IFP2AC(ifp)
# else
#  define WAN_IFP2AC(ifp)		((struct arpcom*)ifp)
# endif
# if defined(IFP2ENADDR)
#  define WAN_IFP2ENADDR(ifp)		IFP2ENADDR(ifp)
# else
#  if defined(IF_LLADDR)
#   define WAN_IFP2ENADDR(ifp)		IF_LLADDR(ifp)
#  else
#   define WAN_IFP2ENADDR(ifp)		(WAN_IFP2AC(ifp))->ac_enaddr
#  endif
# endif
# if  (__FreeBSD_version < 600000)
#  define IFALLOC(if_type)		wan_malloc(sizeof(struct ifnet))
#  define IFFREE(ifp)			wan_free(ifp)
# else
#  define IFALLOC(if_type)		if_alloc(if_type)
#  define IFFREE(ifp)			if_free(ifp)
# endif
# define IFATTACH(ifp)			if_attach(ifp)
# define IFDETACH(ifp)			if_detach(ifp)
# if (__FreeBSD_version < 410000)
#  define ETHER_IFATTACH(ifp,enaddr)	if_attach(ifp); ether_ifattach(ifp)
#  define ETHER_IFDETACH(ifp)		if_detach(ifp)
# elif (__FreeBSD_version > 500000)
#  define ETHER_IFATTACH(ifp,enaddr)	ether_ifattach(ifp, enaddr)
#  define ETHER_IFDETACH(ifp)		ether_ifdetach(ifp)
# else
#  define ETHER_IFATTACH(ifp,enaddr)	ether_ifattach(ifp, ETHER_BPF_SUPPORTED)
#  define ETHER_IFDETACH(ifp)		ether_ifdetach(ifp, ETHER_BPF_SUPPORTED)
# endif
# define ETHER_IOCTL(ifp, cmd, data)	ether_ioctl(ifp, cmd, data)

# define WAN_BPF_DIR_IN		(1<<0)
# define WAN_BPF_DIR_OUT		(1<<1)
# if (__FreeBSD_version > 500000)
#  define WAN_BPF_REPORT(dev,m,d)	bpf_mtap((dev)->if_bpf, (m))
# else
#  define WAN_BPF_REPORT(dev,m,d)	bpf_mtap((dev), (m))
# endif

# define WAN_CDEV_MAJOR			139
# define WAN_MOD_LOAD			MOD_LOAD
# define WAN_MOD_UNLOAD			MOD_UNLOAD
#if 1
# define wan_clear_bit(bit, name) atomic_clear_int((u_int*)(name),(1<<(bit)))
# define wan_set_bit(bit, name)	  atomic_set_int((u_int*)(name),(1<<(bit)))
static __inline int wan_test_bit(int bit, void *name)
{
	int ret;
	//mtx_lock(&Giant);
	ret = atomic_load_acq_int((u_int*)(name)) & (1<<bit);
	//mtx_unlock(&Giant);
	return ret;
}
static __inline int wan_test_and_set_bit(int bit, void *name)
{
	int ret;
	//mtx_lock(&Giant);
	ret = atomic_load_acq_int((u_int*)(name)) & (1<<bit);
	if (!ret){
		wan_set_bit(bit, name);
	}
	//mtx_unlock(&Giant);
	return ret;
}
static __inline int wan_clear_and_set_bit(int bit, void *name)
{
	int ret;
	//mtx_lock(&Giant);
	ret = atomic_load_acq_int((u_int*)(name)) & (1<<bit);
	if (ret){
		wan_clear_bit(bit, name);
	}
	//mtx_unlock(&Giant);
	return ret;
}
#else
# define wan_clear_bit(bit, name)	clear_bit(bit, (name))
# define wan_test_bit(bit, name)	test_bit(bit, (name))
# define wan_set_bit(bit, name)		set_bit(bit, (name))
# define wan_test_and_set_bit(bit, name)	\
		(test_bit((bit), (name)) || set_bit((bit),(name)))
# define wan_test_and_clear_bit(bit, name)	\
		(test_bit((bit), (name)) && clear_bit((bit),(name)))
#endif
# define wan_atomic_set(name, val)	*(int*)(name) = val
# define wan_atomic_read(name)		*(int*)(name)
# define wan_atomic_inc(name)		atomic_add_int((name), 1)
# define wan_atomic_dec(name)		atomic_subtract_int((name), 1)
# define wan_atomic_add(a,name)		atomic_add_int((name), a)
# define wan_atomic_sub(a,name)		atomic_subtract_int((name), a)

# define WAN_IRQ_CALL(fn,args,ret)	fn args
# define WAN_IRQ_RETURN(a)		return
# define WAN_IRQ_RETVAL			void
# define WAN_IRQ_RETVAL_DECL(ret)	
# define WAN_IRQ_RETVAL_SET(ret,v)
# define WAN_IRQ_HANDLED		1
# define WAN_IRQ_NONE			0

# define WAN_PKTATTR_DECL(pktattr)

# define WAN_RTREQUEST(req,dst,gw,nm,fl,rte,tid)	rtrequest(req,dst,gw,nm,fl,rte)

# define WAN_TAILQ_FIRST(ifp)		TAILQ_FIRST(&ifp->if_addrhead)
# define WAN_TAILQ_NEXT(ifa)		TAILQ_NEXT(ifa, ifa_link)

# if (__FreeBSD_version < 410000)
#  define WAN_TASKLET_INIT(task, priority, func, arg)	\
	(task)->running = 0;				\
	(task)->task_func = func; (task)->data = arg
#  define WAN_TASKLET_SCHEDULE(task)			\
	if (!wan_test_bit(0, &(task)->running)){	\
		wan_set_bit(0, &(task)->running);	\
		(task)->task_func((task)->data, 0);	\
	}
#  define __WAN_TASKLET_SCHEDULE(task) WAN_TASKLET_SCHEDULE(task)
#  define WAN_TASKLET_RUNNING(task)					\
		wan_test_bit(0, &(task)->running)
#  define WAN_TASKLET_END(task)	wan_clear_bit(0, &(task)->running)
#  define WAN_TASKLET_KILL(task)
# else
#  define WAN_TASKLET_INIT(task, priority, func, arg)	\
	(task)->running = 0;				\
	TASK_INIT(&(task)->task_id, priority, func, (void*)arg)
#  define WAN_TASKLET_SCHEDULE(task)					\
	if (!wan_test_bit(0, &(task)->running)){			\
		wan_set_bit(0, &(task)->running);			\
		taskqueue_enqueue(taskqueue_swi_giant, &(task)->task_id);\
		taskqueue_run(taskqueue_swi_giant);			\
	}
#  define __WAN_TASKLET_SCHEDULE(task) WAN_TASKLET_SCHEDULE(task)
#  define WAN_TASKLET_RUNNING(task)					\
		wan_test_bit(0, &(task)->running)
#  define WAN_TASKLET_RUN(task)					\
		taskqueue_run(taskqueue_swi_giant);
#  define WAN_TASKLET_END(task)	wan_clear_bit(0, &(task)->running)
#  define WAN_TASKLET_KILL(task)
# endif

# if defined(__FreeBSD__)
#  define WAN_TASKQUEUE_CREATE(name,flags,fn,arg)		\
		taskqueue_create_fast((name),(flags),(fn),(arg))
#  define WAN_TASKQUEUE_START(tq,cnt,pri,arg...)		\
		taskqueue_start_threads((tq),(cnt),(pri),##arg)
#  define WAN_TASKQUEUE_SCHEDULE(tq,task)				\
		taskqueue_enqueue((tq),&(task)->task_id);
#  define WAN_TASKQUEUE_RUN(tq)					\
		taskqueue_run((tq))
#  define WAN_TASKQUEUE_DRAIN(tq,task)				\
		taskqueue_drain((tq),&(task)->task_id);
#  define WAN_TASKQUEUE_FREE(tq)				\
		taskqueue_free((tq))
# endif
# if (__FreeBSD_version < 410000)
#  define WAN_TASKQ_INIT(task, priority, func, arg)	\
		(task)->tfunc = func; task->data = arg
# else
#  define WAN_TASKQ_INIT(task, priority, func, arg)	\
		TASK_INIT(&task->tqueue, priority, func, arg)
# endif
# if (__FreeBSD_version >= 410000)
#  define WAN_IS_TASKQ_SCHEDULE
/* ALEX Oct 15, 2007 change taskq priority queue to taskqueue_swi */
#  define WAN_TASKQ_SCHEDULE(task)			\
	taskqueue_enqueue(taskqueue_swi_giant, &task->tqueue);\
	taskqueue_run(taskqueue_swi_giant)
# else
#  define WAN_IS_TASKQ_SCHEDULE
#  define WAN_TASKQ_SCHEDULE(task)			\
	task->tfunc(task->data, 0)
# endif

typedef struct arpcom	wan_ethercom_t;
# if (__FreeBSD_version > 500000)
typedef struct thread*	wan_proc_ptr_t;
# else
typedef struct proc*	wan_proc_ptr_t;
# endif
# if (__FreeBSD_version >= 503000)
typedef struct cdev*	wan_cdev_t;
# else
typedef dev_t		wan_cdev_t;
# endif
typedef d_open_t	wan_d_open_t;
typedef d_close_t	wan_d_close_t;
typedef d_ioctl_t	wan_d_ioctl_t;

#elif defined(__OpenBSD__)
/************************* O P E N B S D *************************************/
# define IFF_DRV_OACTIVE		IFF_OACTIVE
# define IFF_DRV_RUNNING		IFF_RUNNING
# define WAN_IFP2AC(ifp)		((struct arpcom*)ifp)
# define WAN_IFP2ENADDR(ifp)		(WAN_IFP2AC(ifp))->ac_enaddr
# define ETHER_IFATTACH(ifp, enaddr)	if_attach(ifp); ether_ifattach(ifp)
# define ETHER_IFDETACH(ifp)		ether_ifdetach(ifp); if_detach(ifp)
# define ETHER_IOCTL(ifp, cmd, data)	ether_ioctl(ifp, WAN_IFP2AC(ifp), cmd, data)
# define IFALLOC(if_type)		wan_malloc(sizeof(struct ifnet))
# define IFFREE(ifp)			wan_free(ifp)
# define IFATTACH(ifp)			if_attach(ifp);if_alloc_sadl(ifp)
# if (OpenBSD > 200311)
#  define IFDETACH(ifp)			if_detach(ifp)
# else
#  define IFDETACH(ifp)			if_free_sadl(ifp); if_detach(ifp)
# endif

# if (OpenBSD < 200611)
#  define WAN_BPF_DIR_IN		(1<<0)
#  define WAN_BPF_DIR_OUT		(1<<1)
#  define WAN_BPF_REPORT(dev,m,d)	bpf_mtap((dev)->if_bpf, (m));
# else
#  define WAN_BPF_DIR_IN		BPF_DIRECTION_IN
#  define WAN_BPF_DIR_OUT		BPF_DIRECTION_OUT
#  define WAN_BPF_REPORT(dev,m,dir)					\
	if (dir == WAN_BPF_DIR_IN){					\
		bpf_mtap((dev)->if_bpf, (m), BPF_DIRECTION_IN);		\
	}else{								\
		bpf_mtap((dev)->if_bpf, (m), BPF_DIRECTION_OUT);	\
	}
# endif

# define WAN_MOD_LOAD			LKM_E_LOAD
# define WAN_MOD_UNLOAD			LKM_E_UNLOAD
# define wan_clear_bit(bit, name)	bit_clear((unsigned char*)(name), bit)
# define wan_set_bit(bit, name)		bit_set((unsigned char*)(name), bit)
# define wan_test_bit(bit, name)	bit_test((unsigned char*)(name), bit)
# define wan_test_and_set_bit(bit, name)		\
    	(bit_test((unsigned char*)(name),(bit)) || 	\
		!bit_set((unsigned char*)(name),(bit)))
# define wan_test_and_clear_bit(bit, name)		\
    	(bit_test((unsigned char*)(name),(bit)) && 	\
		!bit_clear((unsigned char*)(name),(bit)))
# define wan_atomic_set(name, val)	*(int*)(name) = val
# define wan_atomic_set_int(name, val)	*(unsigned int*)(name) = val
# define wan_atomic_read(name)		*(int*)(name)
# define wan_atomic_inc(name)		*(int*)(name) += 1
# define wan_atomic_dec(name)		*(int*)(name) -= 1 
# define wan_atomic_add(a,name)		*(int*)(name) += a
# define wan_atomic_sub(a,name)		*(int*)(name) -= a 

/* Unsafe sprintf and vsprintf function removed from the kernel */
# define sprintf(a,b...)		snprintf(a, sizeof(a), ##b)
# define vsprintf(a,b...)		vsnprintf(a, sizeof(a), ##b)

# define WAN_IRQ_CALL(fn,args,ret)	ret = fn args
# define WAN_IRQ_RETURN(a)		return (a)
# define WAN_IRQ_RETVAL			int
# define WAN_IRQ_RETVAL_SET(ret, val)	ret = val
# define WAN_IRQ_RETVAL_DECL(ret)	int ret
# define WAN_IRQ_NONE			0
# define WAN_IRQ_HANDLED		1

# if (OpenBSD >= 200611)
#  define WAN_RTREQUEST(req,dst,gw,nm,fl,rte,tid)	rtrequest(req,dst,gw,nm,fl,rte,tid)
# else
#  define WAN_RTREQUEST(req,dst,gw,nm,fl,rte,tid)	rtrequest(req,dst,gw,nm,fl,rte)
# endif

# define WAN_TAILQ_FIRST(ifp)		TAILQ_FIRST(&ifp->if_addrlist)
# define WAN_TAILQ_NEXT(ifa)		TAILQ_NEXT(ifa, ifa_list)

# define WAN_TASKLET_INIT(task, priority, func, arg)	\
	(task)->running = 0;				\
	(task)->task_func = func; (task)->data = arg
# define WAN_TASKLET_SCHEDULE(task)					\
	if (!wan_test_bit(0, &(task)->running)){			\
		wan_set_bit(0, &(task)->running);			\
		(task)->task_func((task)->data, 0);			\
	}
# define __WAN_TASKLET_SCHEDULE(task) WAN_TASKLET_SCHEDULE(task)
# define WAN_TASKLET_RUNNING(task)					\
		wan_test_bit(0, &(task)->running)
# define WAN_TASKLET_END(task)	wan_clear_bit(0, &(task)->running)
# define WAN_TASKLET_KILL(task)

# define WAN_TASKQ_INIT(task, priority, func, arg)	\
		(task)->tfunc = func; task->data = arg
# define WAN_IS_TASKQ_SCHEDULE
# define WAN_TASKQ_SCHEDULE(task)			\
	task->tfunc(task->data, 0)

typedef struct proc*	wan_proc_ptr_t;
typedef struct arpcom	wan_ethercom_t;
typedef dev_t	wan_cdev_t;
typedef int wan_d_open_t (wan_cdev_t, int, int, struct proc*);
typedef int wan_d_close_t (wan_cdev_t, int, int, struct proc*);
typedef int wan_d_ioctl_t (wan_cdev_t, u_long, caddr_t, int, struct proc*);

#elif defined(__NetBSD__)
/************************* N E T B S D *************************************/
# define IFF_DRV_OACTIVE		IFF_OACTIVE
# define IFF_DRV_RUNNING		IFF_RUNNING
# define WAN_IFP2AC(ifp)		((struct ethercom*)ifp)
# define WAN_IFP2ENADDR(ifp)		(WAN_IFP2AC(ifp))->ac_enaddr
# define ETHER_IFATTACH(ifp,enaddr)	if_attach(ifp);ether_ifattach(ifp,enaddr)
# define ETHER_IFDETACH(ifp)		ether_ifdetach(ifp); if_detach(ifp)
# define ETHER_IOCTL(ifp, cmd, data)	ether_ioctl(ifp, cmd, data)
# define IFALLOC(if_type)		wan_malloc(sizeof(struct ifnet))
# define IFFREE(ifp)			wan_free(ifp)
# define IFATTACH(ifp)			if_attach(ifp);if_alloc_sadl(ifp)
# define IFDETACH(ifp)			if_free_sadl(ifp); if_detach(ifp)

#  define WAN_BPF_DIR_IN		(1<<0)
#  define WAN_BPF_DIR_OUT		(1<<1)
#  define WAN_BPF_REPORT(dev,m,d)	bpf_mtap((dev)->if_bpf, (m));

# define WAN_MOD_LOAD			LKM_E_LOAD
# define WAN_MOD_UNLOAD			LKM_E_UNLOAD
# define wan_clear_bit(bit, name)	bit_clear((unsigned char*)(name), bit)
# define wan_set_bit(bit, name)		bit_set((unsigned char*)(name), bit)
# define wan_test_bit(bit, name)	bit_test((unsigned char*)(name), bit)
# define wan_test_and_set_bit(bit, name)		\
    	(bit_test((unsigned char*)(name),(bit)) || 	\
		!bit_set((unsigned char*)(name),(bit)))
# define wan_test_and_clear_bit(bit, name)		\
    	(bit_test((unsigned char*)(name),(bit)) && 	\
		!bit_clear((unsigned char*)(name),(bit)))
# define wan_atomic_set(name, val)	*(int*)(name) = val
# define wan_atomic_set_int(name, val)	*(unsigned int*)(name) = val
# define wan_atomic_read(name)		*(int*)(name)
# define wan_atomic_inc(name)		*(int*)(name) += 1
# define wan_atomic_dec(name)		*(int*)(name) -= 1 
# define wan_atomic_add(a,name)		*(int*)(name) += a
# define wan_atomic_sub(a,name)		*(int*)(name) -= a 

/* Unsafe sprintf and vsprintf function removed from the kernel */

# define WAN_IRQ_CALL(fn,args,ret)	fn args
# define WAN_IRQ_RETURN(a)
# define WAN_IRQ_RETVAL			void
# define WAN_IRQ_RETVAL_SET(ret, val)
# define WAN_IRQ_RETVAL_DECL(ret)
# define WAN_IRQ_NONE			0
# define WAN_IRQ_HANDLED		1

# define WAN_PKTATTR_DECL(pktattr)		struct altq_pktattr	pktattr

# define WAN_RTREQUEST(req,dst,gw,nm,fl,rte,tid)	rtrequest(req,dst,gw,nm,fl,rte)

# define WAN_TAILQ_FIRST(ifp)		TAILQ_FIRST(&ifp->if_addrlist)
# define WAN_TAILQ_NEXT(ifa)		TAILQ_NEXT(ifa, ifa_list)

# define WAN_TASKLET_INIT(task, priority, func, arg)	\
	(task)->running = 0;				\
	(task)->task_func = func; (task)->data = arg
# define WAN_TASKLET_SCHEDULE(task)					\
	if (!wan_test_bit(0, &(task)->running)){			\
		wan_set_bit(0, &(task)->running);			\
		(task)->task_func((task)->data, 0);			\
	}
# define __WAN_TASKLET_SCHEDULE(task) WAN_TASKLET_SCHEDULE(task)
# define WAN_TASKLET_RUNNING(task)					\
		wan_test_bit(0, &(task)->running)
# define WAN_TASKLET_END(task)	wan_clear_bit(0, &(task)->running)
# define WAN_TASKLET_KILL(task)

# define WAN_TASKQ_INIT(task, priority, func, arg)	\
		(task)->tfunc = func; task->data = arg
# define WAN_IS_TASKQ_SCHEDULE
# define WAN_TASKQ_SCHEDULE(task)			\
	task->tfunc(task->data, 0)

typedef struct proc*	wan_proc_ptr_t;
typedef struct ethercom	wan_ethercom_t;
typedef dev_t	wan_cdev_t;
typedef int wan_d_open_t (wan_cdev_t, int, int, struct proc*);
typedef int wan_d_close_t (wan_cdev_t, int, int, struct proc*);
typedef int wan_d_ioctl_t (wan_cdev_t, u_long, caddr_t, int, struct proc*);
#else
# error "IFATTACH/IFDETACH/ETHER_IFDETACH macros doesn't supported yet!"
#endif
/***************************************************************************/

/* IFQ macros */
#ifndef IFQ_SET_READY
# define WAN_IFQ_SET_READY(ifq)		/* nothing */
#else
# define WAN_IFQ_SET_READY		IFQ_SET_READY
#endif
#ifndef IFQ_IS_EMPTY
# define WAN_IFQ_IS_EMPTY(ifq)		((ifq)->ifq_len == 0)
#else
# define WAN_IFQ_IS_EMPTY		IFQ_IS_EMPTY
#endif
#ifndef IFQ_INC_LEN
# define WAN_IFQ_INC_LEN(ifq)		((ifq)->ifq_len++)
#else
# define WAN_IFQ_INC_LEN		IFQ_INC_LEN
#endif
#ifndef IFQ_DEC_LEN
# define WAN_IFQ_DEC_LEN(ifq)		(--(ifq)->ifq_len)
#else
# define WAN_IFQ_DEC_LEN		IFQ_DEC_LEN
#endif
#ifndef IFQ_INC_DROPS
# define WAN_IFQ_INC_DROPS(ifq)		((ifq)->ifq_drops++)
#else
# define WAN_IFQ_INC_DROPS		IFQ_INC_DROPS
#endif
#ifndef IFQ_SET_MAXLEN
# define WAN_IFQ_SET_MAXLEN(ifq, len)	((ifq)->ifq_maxlen = (len))
#else
# define WAN_IFQ_SET_MAXLEN		IFQ_SET_MAXLEN
#endif
#if (__FreeBSD_version > 500000)
# define IF_DROP	_IF_DROP
# define IF_QFULL	_IF_QFULL
#endif
#ifndef IFQ_PURGE
# if (__FreeBSD_version > 500000)
#  define WAN_IFQ_PURGE(ifq)	IF_DRAIN((ifq))
# else
#  define WAN_IFQ_PURGE(ifq)					\
	while (1) {						\
		struct mbuf *m0;				\
		IF_DEQUEUE((ifq), m0);				\
		if (m0 == NULL)					\
			break;					\
		else						\
			m_freem(m0);				\
	}
# endif
#else
# define WAN_IFQ_PURGE				IFQ_PURGE
#endif
# define WAN_IFQ_DMA_PURGE			WAN_IFQ_PURGE
#ifndef IFQ_ENQUEUE
# if (__FreeBSD_version > 500000)
#  define WAN_IFQ_ENQUEUE(ifq, m, pattr, err)	IF_ENQUEUE((ifq), (m))
# else
#  define WAN_IFQ_ENQUEUE(ifq, m, pattr, err)			\
	do {							\
		if (IF_QFULL((ifq))) {				\
			m_freem((m));				\
			(err) = ENOBUFS;			\
		} else {					\
			IF_ENQUEUE((ifq), (m));			\
			(err) = 0;				\
		}						\
		if ((err))					\
			(ifq)->ifq_drops++;			\
	} while (0)
# endif
#else
# if (__FreeBSD_version > 503000)
#  define WAN_IFQ_ENQUEUE(ifq, m, pattr, err)	IFQ_ENQUEUE((ifq),(m),(err))
# else
#  define WAN_IFQ_ENQUEUE			IFQ_ENQUEUE
# endif
#endif
#ifndef IFQ_DEQUEUE
# define WAN_IFQ_DEQUEUE(ifq, m)		IF_DEQUEUE((ifq), (m))
#else
# define WAN_IFQ_DEQUEUE			IFQ_DEQUEUE
#endif
#ifndef IFQ_POLL
# define WAN_IFQ_POLL(ifq, m)			((m) = (ifq)->ifq_head)
#else
# define WAN_IFQ_POLL				IFQ_POLL
#endif
#ifndef IFQ_CLASSIFY
/*struct altq_pktattr { void* attr; };*/
# define WAN_IFQ_CLASSIFY(ifq, m, sa_family, pktattr) /*(pktattr)->attr = NULL*/
#else
# define WAN_IFQ_CLASSIFY			IFQ_CLASSIFY
#endif
#if (__FreeBSD_version > 500000)
# define WAN_IFQ_INIT(ifq, max_pkt)		\
		IFQ_SET_MAXLEN((ifq), (max_pkt)?(max_pkt):IFQ_MAXLEN);	\
		mtx_init(&(ifq)->ifq_mtx, "ifq", "wan", MTX_DEF);	\
 		(ifq)->ifq_len = 0;
# define WAN_IFQ_DESTROY(ifq) 					\
		mtx_destroy(&(ifq)->ifq_mtx);
#else
# define WAN_IFQ_INIT(ifq, max_pkt)		\
		IFQ_SET_MAXLEN((ifq), max_pkt);	\
 		(ifq)->ifq_len = 0;
# define WAN_IFQ_DESTROY(ifq)
#endif
#define WAN_IFQ_LEN(ifq)	((ifq)->ifq_len)

/* BPF macros */
#if defined(__NetBSD__) || defined(__FreeBSD__) || defined(__OpenBSD__)
# ifdef NBPFILTER
#  if defined(__FreeBSD__)
#   define BPF_ATTACH(ifp,type)	bpfattach(ifp, type, 4)
#  elif defined(__OpenBSD__)
#   define BPF_ATTACH(ifp,type)	bpfattach(&ifp->if_bpf, ifp, DLT_NULL, 4)
#  elif defined(__NetBSD__)
#   define BPF_ATTACH(ifp,type)	bpfattach(ifp, type, 4)
#  else
#   error "BPF_ATTACH macro doesn't supported yet!"
#  endif
# else
#  define BPF_ATTACH(ifp, type)
# endif
#endif


#endif /* WAN_KERNEL */

#endif /* __WANPIPE_KERNEL_H */
