#!/bin/sh
# support script for binding interfaces
#   uses netgraph
#
#----------------------------------------------------------------------
# Aug 23, 2007		Alex	Reading meta conf file location from
#				/etc/rc.conf
# initial version: 	jmp_zer0  01.31.2007

print_usage()
{
	echo "$0 command parent_if <child1> <child2> ... <childn>          "
        echo "      where:"
        echo "         command    specifies the requested operation        "
        echo "                                                             "
        echo "                    bond        -  bond the interfaces       "
	echo "                                    specified                "
	echo "                    bond_some   -  bond child interfaces     "
        echo "                    debond      -  remove entire bonding     "
	echo "                    debond_some -  remove a child interface  "
	echo "			  new_bond_iface - create new virtual      "
	echo "					    interface for bonding  "
	echo "                    setip       -  set the ip address of     "
	echo "			 		  the parent interface     "
	echo "                                                             "
        echo "         parent_if  specifies the parent interface           "
        echo "                             for this bonding                "
        echo "                                                             "
        echo "         childx            specifies the child interfaces    "
        echo "                                                             "
}

# create ng_iface interface as virtual interface in bonding
#  - interface name is auto-assigned by ng_iface - ng0, ng1, ..., ngx 
#     so no parameter is required
create_parent_iface()
{
	if [ -z $1 ]; then
		debug_output "${NGCTL_CMD} mkpeer . iface hook inet"
		${NGCTL_CMD} mkpeer . iface hook inet
	fi
	
	if [ "`${IFCONFIG_CMD} $1 2>/dev/null | wc -l`" -eq "0" ]; then
		debug_output "${NGCTL_CMD} mkpeer . iface hook inet"
		${NGCTL_CMD} mkpeer . iface hook inet
	fi
}

# destroy ng_iface interface
#  (note: "ngctl shutdown" deletes the interface, but this is provided
#          for brute-force mechanism)
#
#   e.g. destroy_bond_iface <iface_name>
destroy_bond_iface()
{
	debug_output "${NGCTL_CMD} shutdown ${1}:"
	${NGCTL_CMD} shutdown ${1}:
}

# setup ng_one2many node and parent interface
setup_ngone_node()
{
	debug_output "${NGCTL_CMD} mkpeer ${1}: one2many inet one"
	${NGCTL_CMD} mkpeer ${1}: one2many inet one
}

# destroy ng_one2many node and parent interface
destroy_ngone_node()
{
	if [ "`${NGCTL_CMD} show ${1}:inet 2>/dev/null | wc -l`" -gt "0" ]; then
		debug_output "${NGCTL_CMD} shutdown ${1}:inet"
		${NGCTL_CMD} shutdown ${1}:inet
	fi
}

# add a child iface for bonding
add_child_bond()
{
	parent_iface=$1
	child_iface=$2
	idx_no=$3

	debug_output "${NGCTL_CMD} connect ${child_iface}: ${parent_if}:inet lower many${idx_no}"
	${NGCTL_CMD} connect ${child_iface}: ${parent_if}:inet lower many${idx_no}
}

# remove a child from bonded iface
rm_child_bond()
{
	child_iface=$1

	debug_output "${NGCTL_CMD} rmhook ${child_iface}: lower"
	${NGCTL_CMD} rmhook ${child_iface}: lower
}

# set to promiscous mode
set_promisc_mode()
{
	iface=$1
	val=$2

	debug_output "${NGCTL_CMD} msg ${iface}: setpromisc ${val}"
	${NGCTL_CMD} msg ${iface}: setpromisc ${val}
}

# unset autosrc mode
set_autosrc_mode()
{
	iface=$1
	val=$2

	debug_output "${NGCTL_CMD} msg ${iface}: setautosrc ${val}"
	${NGCTL_CMD} msg ${iface}: setautosrc ${val}
}

# clear ip address of interface
clear_ip_addr()
{
	if [ "`${IFCONFIG_CMD} $1 2>/dev/null | wc -l`" -ge "2" ]; then
		${IFCONFIG_CMD} $1 delete
	fi
}

# check for valid commands
check_command_valid()
{
	case $1 in
		bond|bond_some|debond|debond_some|new_bond_iface|setip)
			;;
		*)
			echo "Invalid command: $1!"
			print_usage
			exit 2
			;;
	esac
}

# debug print (set by vflag)
debug_output()
{
	if [ "${vflag}" -eq "1" ]; then
		echo $1
	fi
}

# Read meta config file
read_meta_conf()
{
	WAN_HOME=/usr/local/etc/wanpipe
	META_CONF=$WAN_HOME/wanrouter.rc
	wanrouter_rc_file=""
	if [ -r /etc/rc.conf ]; then
		. /etc/rc.conf
	fi
	if [ -n "$wanrouter_rc_file" ]; then
		META_CONF=${wanrouter_rc_file}
	fi

	echo "${META_CONF}" 
	. ${META_CONF}
	return 0
}

#############################################################################
# start here

read_meta_conf

NGCTL_CMD=/usr/sbin/ngctl
IFCONFIG_CMD=/sbin/ifconfig
ROUTE_CMD=/sbin/route
WANROUTER_CMD=${WAN_BIN_DIR}/wanrouter

# sanity check - requires at least the "cmd" paremeter
if [ $# -lt 1 ]; then
	print_usage
	exit 1
fi

# get command and parent interface from parameter list
op_cmd=$1
shift

# set default global vars
vflag=0

# check if command type is valid
check_command_valid $op_cmd

# get parameters
args=`getopt vg:p: $*`
if [ $? -ne 0 ]; then
	print_usage
	exit 2
fi
set -- $args
for i
do
	case "$i"
	in
		# verbose flag
		-v)
			vflag=1;
			shift;;

		# virtual/parent interface
		-p)
			parent_if="$2";
			shift;shift;;

		--)
			shift;
			break;;
	esac
done

# if parent/virtual interface is not set, defaults to ng0
if [ -z $parent_if ]; then
	parent_if="ng0"
fi

# switch for command type
case $op_cmd in
	# set bonding
	bond)
		create_parent_iface $parent_if
		destroy_ngone_node $parent_if
		setup_ngone_node $parent_if
		idx=0
		for i in $@
		do
			clear_ip_addr $i
			set_promisc_mode $i 1
			set_autosrc_mode $i 0
			add_child_bond $parent_if $i $idx
			link_str="$link_str""1 "
			idx=`expr $idx + 1`
		done

		# set ng_one2many node configuration
		${NGCTL_CMD} msg ${parent_if}:inet setconfig \
			"{ xmitAlg=1 failAlg=2 enabledLinks=[ ${link_str} ] }"
		debug_output "${NGCTL_CMD} msg ${parent_if}:inet setconfig \
			'{ xmitAlg=1 failAlg=2 enabledLinks=[ ${link_str} ] }'"
		;;

	# add child interfaces to bond
	#  - assumes ng_one2many node and parent hooks are setup
	bond_some)
		echo "Not supported yet!"
		echo "Difficult to determine index of many<x> hook!"
		exit 1
		;;

	# remove one or more child nodes from bond
	debond_some)
		for i in $@
		do
			rm_child_bond $i
		done
		;;

	# destroy bonded interface and restore wancfg-set ip addresses
	#  also restore default route with supplied default gw
	debond)
		destroy_ngone_node $parent_if
		destroy_bond_iface $parent_if
		$WANROUTER_CMD start_ip
		# set specified default gw
		if [ ! -z $1 ]; then 
			${ROUTE_CMD} add default $1
		fi
		;;

	# create ng_iface as parent of bonding
	new_bond_iface)
		create_parent_iface
		;;

	# set the ip of the parent interface and set default route 
	setip)
		debug_output "${IFCONFIG_CMD} $parent_if $1 $2 netmask $3"
		${IFCONFIG_CMD} $parent_if $1 $2 netmask $3
		debug_output "${ROUTE_CMD} add default $2"
		${ROUTE_CMD} add default $2
		;;

	*)
		print_usage
		exit 1
		;;
esac

exit 0
